/***********************************************************************
 * File: UTM.c
 * (Extracted from
 * "$Id: UTM.c,v 1.5 2009/05/30 17:01:29 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source code implements a test program for buffer_checker.c.
 *  This works as a Universal Turing's machine.
 *  If You like to try this, type the following.  This prints "15".
 *	$ TM_sum.ksh 1 2 3 4 5
 *  TM_sum.ksh is another implements of TM_sum. It executes UTM.
 *
 * Copyright (C) 2004 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

static int debug = /* switch for debugging: 0 means no-debug. */
#		if (defined(DEBUG) && DEBUG > 0) /* { */
			DEBUG
#		else /* } { */
			0
#		endif /* } */
		;

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

/* print usage of this program "Universal Turing's Machine" */
static void usage()
{
	fprintf(stderr
		, "Usage: UTM -p '(status-cnt ((UTM-func1 next-status1) "
				"(UTM-func2 next-status2))...)' "
				"[num...]\n"
		"    a. `status-cnt' is the count of status.\n"
		); fflush(stderr)
		;
	fprintf(stderr,
		"    b. `UTM-func' is the code name of UTM functions.\n"
		"        UTM function codes are defined...\n"
		); fflush(stderr)
		;
	fprintf(stderr,
		"          0) term - terminate immediately\n"
		"          1) noop - no operation\n"
		"          2) wrt0 - set \"0\" the current cell\n"
		"          3) wrt1 - set \"1\" the current cell\n"
		"          4) next - forward the cell\n"
		"          5) prev - backward the cell\n"
		"          6) wrt0_term - wrt0 and term\n"
		"          7)   <<same as '2) wrt0'>>\n"
		"          8)   <<same as '2) wrt0'>>\n"
		"          9)   <<same as '3) wrt1'>>\n"
		"         10) wrt0_next - '2) wrt0' and '4) next'\n"
		); fflush(stderr)
		;
	fprintf(stderr,
		"         11) wrt0_prev - '2) wrt0' and '5) prev'\n"
		"         12) wrt1_term - '3) wrt1' and '0) term'\n"
		"         13)   <<same as '3) wrt1'>>\n"
		"         14)   <<same as '2) wrt0'>>\n"
		"         15)   <<same as '3) wrt1'>>\n"
		"         16) wrt1_next - '3) wrt1' and '4) next'\n"
		"         17) wrt1_prev - '3) wrt1' and '5) prev'\n"
		"         18) next_term - '4) next' and '0) term'\n"
		"         19)   <<same as '4) next'>>\n"
		"         20) next_wrt0 - '4) next' and '2) wrt0'\n"
		); fflush(stderr)
		;
	fprintf(stderr,
		"         21) next_wrt1 - '4) next' and '3) wrt1'\n"
		"         22) next_next - '4) next' and '4) next'\n"
		"         23)   <<same as '1) noop'>>\n"
		"         24) prev_term - '5) prev' and '0) term'\n"
		"         25)   <<same as '5) prev'>>\n"
		"         26) prev_wrt0 - '5) prev' and '2) wrt0'\n"
		"         27) prev_wrt1 - '5) prev' and '3) wrt1'\n"
		"         28)   <<same as '1) noop'>>\n"
		"         29) prev_prev - '5) prev' and '5) prev'\n"
		); fflush(stderr)
		;
	fprintf(stderr,
		"    c. `next-status' must be zero or less than "
							"`status-cnt'\n"
		"    d. `num' must be the positive integer.\n"
		); fflush(stderr)
		;
	return;
}


#include <buffer_checker.h> /* include the buffer_checker header file */


/* type-definition of programmed UTM machine structure */
typedef struct utm_machine_t {
#	define CELL_SZ	2	/* size of a cell of tape is 2 bytes */
#	define CELL_0	"0"	/* value of zero-ed cell */
#	define CELL_1	"1"	/* value of one-ed cell */
#	define WRITE_0(CELLP) (strcpy((char *)(CELLP), CELL_0))
				/* write "0" the current cell of tape */
#	define WRITE_1(CELLP) (strcpy((char *)(CELLP), CELL_1))
				/* write "1" the current cell of tape */

	ssize_t		head;	/* the position of tape-head */
#	define	NEXT(INTERNALP)	((INTERNALP)->head += CELL_SZ)
				/* forward the cell */
#	define	PREV(INTERNALP)	((INTERNALP)->head -= CELL_SZ)
				/* backward the cell */

	ssize_t		status_cnt; /* the count of status */
	ssize_t		status;	/*  the current status of UTM */

	Buffer_checker_t *defines; /* status and function pairs */
	ssize_t		*matched_next_status; /* next status for CELL_0 */
	ssize_t		*unmatched_next_status; /* next status for CELL_1 */
} UTM_machine_t;	/* programmed UTM machine structure */

static UTM_machine_t machine; /* programmed UTM machine */


/* For debugging, dump the UTM program */
static void dump_program(UTM_machine_t *machinep)
{
	ssize_t	i;

	if (!debug) return;	/* not debugging mode */

	for (i = 0; i < machinep->status_cnt; ++i) {
		printf("status#%d: fld_offset = %ld, "
				"fldsz = %lu, "
				"user_get_fld_addr = 0x%x, "
				"regex = \"%s\", "
				"preg = %p, "
				"clause_cba[0] = 0x%x, "
				"clause_cba[1] = 0x%x "
				"clause_cba[2] = 0x%x "
				"matched_next_status = %d, "
				"unmatched_next_status = %d.\n"
				,    i, (long)machinep->defines[i].fld_offset
				, (unsigned long)machinep->defines[i].fldsz
				, (unsigned)(machinep->defines[i]
							.user_get_fld_addr)
				,	machinep->defines[i].regex
				, (void *)(machinep->defines[i].preg)
				, (unsigned)(machinep->defines[i].clause_cbs[0])
				, (unsigned)(machinep->defines[i].clause_cbs[1])
				, (unsigned)(machinep->defines[i].clause_cbs[2])
				,	machinep->matched_next_status[i]
				,	machinep->unmatched_next_status[i]
				); fflush(stdout)
				;
	}

	return;
}


/* For debugging, dump the whole cells of the tape */
static void dump_tape(char *tape, ssize_t tape_sz)
{
	ssize_t	i;

	if (debug) {
		for (i = 0; i < tape_sz; i += CELL_SZ) {
			printf("%d-th cell: %s\n", i, tape + i); fflush(stdout);
		}
		printf("tape_sz = %ld.\n", (long)tape_sz); fflush(stdout);
	}

	return;
}


/* For debugging, dump the internal status; tape head, status,...etc. */
static void dump_internal(UTM_machine_t *internal)
{
	if (debug) {
		printf("internal->head = %d, "
				"internal->status_cnt = %d, "
				"internal->status = %d.\n"
				,		internal->head
				,		internal->status_cnt
				,		internal->status
				); fflush(stdout)
				;
	}

	return;
}


/* test is the internal status illegal? */
#define ILLEGAL_INTERNAL_ARG	((internal == NULL)						  \
								|| internal->status > internal->status_cnt\
								)


/* test are the arguments of UTM functions illegal? */
#define ILLEGAL_CALLBACKS_ARGS(FUNCNAME)					\
{		if (((((ILLEGAL_INTERNAL_ARG)						\
		|| (ssize_t)checker_index != (internal->status))	\
		|| buffer == NULL)									\
		|| buffer_sz < CELL_SZ)								\
		|| fldp == NULL) {									\
			return BCR_Unmatched;							\
		}													\
		else if (debug) {									\
			printf("***DBG: %s():\n", (FUNCNAME));			\
			fflush(stdout);									\
			dump_internal(internal);						\
		}													\
}


/* get the address of the current cell of the tape.
 * this returns address of a cell at tape-head.
 */
static void *get_cell_addr(void	 		*buffer
							, ssize_t	fld_offset
							, size_t	*fldszp
							, void	 	*user_cntxt
							)
{
	UTM_machine_t *internal = (UTM_machine_t *)user_cntxt;

	/* test arguments and internal status */
	if ((((buffer == NULL)
	|| fld_offset != 0)
	|| fldszp == NULL)
	|| ILLEGAL_INTERNAL_ARG
	) {	/* !!FATAL!! Illegal arguments or Illegal internal status */
		return	NULL;
	}

	/* returns address of the cell under the tape-head */
	return	(char *)buffer + internal->head;
}


/* get the next status */
static ssize_t get_next_status(char *cellp, ssize_t current_status)
{
	ssize_t	next_status;

	next_status = ((strncmp(cellp, CELL_0, CELL_SZ) == 0)?
						machine.matched_next_status[current_status]
					:	machine.unmatched_next_status[current_status]
					);

	return	next_status;
}

#define IS_NEXT_STATUS_SAME_CURRENT(INTERNAL,NEXT_STATUS)	\
							(INTERNAL->status == NEXT_STATUS)
#define UTM_Detected_Infinite_Loop -1


/* UTM function; TERM */
static int term(size_t	 	checker_index
				, void		*buffer
				, size_t	buffer_sz
				, void		*fldp
				, void		*user_cntxt
				)
{
	UTM_machine_t *internal = (UTM_machine_t *)user_cntxt;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("term");

	internal->status = (ssize_t)get_next_status((char *)fldp
												, internal->status
												);
	return	BCR_Terminate;
}


/* UTM function; NOOP */
static int noop(size_t	 	checker_index
				, void		*buffer
				, size_t	buffer_sz
				, void		*fldp
				, void		*user_cntxt
				)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("noop");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	return	(IS_NEXT_STATUS_SAME_CURRENT(internal, next_status))?
				UTM_Detected_Infinite_Loop
			:	(internal->status = next_status)
			;
}


/* UTM function; WRITE CELL_0 */
static int wrt0(size_t	 	checker_index
				, void		*buffer
				, size_t	buffer_sz
				, void		*fldp
				, void		*user_cntxt
				)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("wrt0");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	if ((IS_NEXT_STATUS_SAME_CURRENT(internal, next_status))
	&& strcmp((char *)fldp, CELL_0) == 0
	) {	return	UTM_Detected_Infinite_Loop;	}

	WRITE_0(fldp);
	internal->status = next_status;
	return	internal->status;
}


/* UTM function; WRITE CELL_1 */
static int wrt1(size_t	 	checker_index
				, void		*buffer
				, size_t	buffer_sz
				, void		*fldp
				, void		*user_cntxt
				)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("wrt1");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	if ((IS_NEXT_STATUS_SAME_CURRENT(internal, next_status))
	&& strcmp((char *)fldp, CELL_1) == 0
	) {	return	UTM_Detected_Infinite_Loop;	}

	WRITE_1(fldp);
	internal->status = next_status;
	return	internal->status;
}


/* UTM function; NEXT */
static int next(size_t	 	checker_index
				, void		*buffer
				, size_t	buffer_sz
				, void		*fldp
				, void		*user_cntxt
				)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("next");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	NEXT(internal);
	internal->status = next_status;
	return	internal->status;
}


/* UTM function; PREV */
static int prev(size_t	 	checker_index
				, void		*buffer
				, size_t	buffer_sz
				, void		*fldp
				, void		*user_cntxt
				)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("prev")

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	PREV(internal);
	internal->status = next_status;
	return	internal->status;
}


/* UTM function; WRITE CELL_0 and TERM */
static int wrt0_term(size_t	 	checker_index
					, void		*buffer
					, size_t	buffer_sz
					, void		*fldp
					, void		*user_cntxt
					)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("wrt0_term");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	WRITE_0(fldp);
	internal->status = next_status;
	return	BCR_Terminate;
}


/* UTM function; WRITE CELL_1 and TERM */
static int wrt1_term(size_t	 	checker_index
					, void		*buffer
					, size_t	buffer_sz
					, void		*fldp
					, void		*user_cntxt
					)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("wrt1_term");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	WRITE_1(fldp);
	internal->status = next_status;
	return	BCR_Terminate;
}


/* UTM function; NEXT and TERM */
static int next_term(size_t	 	checker_index
					, void		*buffer
					, size_t	buffer_sz
					, void		*fldp
					, void		*user_cntxt
					)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("next_term");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	NEXT(internal);
	internal->status = next_status;
	return	BCR_Terminate;
}


/* UTM function; PREV and TERM */
static int prev_term(size_t	 	checker_index
					, void		*buffer
					, size_t	buffer_sz
					, void		*fldp
					, void		*user_cntxt
					)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("prev_term");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	PREV(internal);
	internal->status = next_status;
	return	BCR_Terminate;
}


/* UTM function; WRITE CELL_0 and NEXT */
static int wrt0_next(size_t	 	checker_index
					, void		*buffer
					, size_t	buffer_sz
					, void		*fldp
					, void		*user_cntxt
					)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("wrt0_next");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	WRITE_0(fldp);
	NEXT(internal);
	internal->status = next_status;
	return	internal->status;
}


/* UTM function; WRITE CELL_0 and PREV */
static int wrt0_prev(size_t	 	checker_index
					, void		*buffer
					, size_t	buffer_sz
					, void		*fldp
					, void		*user_cntxt
					)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("wrt0_prev");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	WRITE_0(fldp);
	PREV(internal);
	internal->status = next_status;
	return	internal->status;
}

/* UTM function; WRITE CELL_1 and NEXT */
static int wrt1_next(size_t	 	checker_index
					, void		*buffer
					, size_t	buffer_sz
					, void		*fldp
					, void		*user_cntxt
					)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("wrt1_next");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	WRITE_1(fldp);
	NEXT(internal);
	internal->status = next_status;
	return	internal->status;
}


/* UTM function; WRITE CELL_1 and PREV */
static int wrt1_prev(size_t	 	checker_index
					, void		*buffer
					, size_t	buffer_sz
					, void		*fldp
					, void		*user_cntxt
					)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("wrt1_prev");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	WRITE_1(fldp);
	PREV(internal);
	internal->status = next_status;
	return	internal->status;
}


/* UTM function; NEXT and WRITE CELL_0 */
static int next_wrt0(size_t	 	checker_index
					, void		*buffer
					, size_t	buffer_sz
					, void		*fldp
					, void		*user_cntxt
					)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("next_wrt0");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	NEXT(internal);
	WRITE_0(fldp);
	internal->status = next_status;
	return	internal->status;
}


/* UTM function; NEXT and WRITE CELL_1 */
static int next_wrt1(size_t	 	checker_index
					, void		*buffer
					, size_t	buffer_sz
					, void		*fldp
					, void		*user_cntxt
					)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("next_wrt1");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	NEXT(internal);
	WRITE_1(fldp);
	internal->status = next_status;
	return	internal->status;
}


/* UTM function; NEXT and NEXT */
static int next_next(size_t	 	checker_index
					, void		*buffer
					, size_t	buffer_sz
					, void		*fldp
					, void		*user_cntxt
					)
{
	UTM_machine_t	*internal = (UTM_machine_t *)user_cntxt;
	ssize_t			next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("next_next");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	NEXT(internal);
	NEXT(internal);
	internal->status = next_status;
	return	internal->status;
}


/* UTM function; PREV and WRITE CELL_0 */
static int prev_wrt0(size_t	 	checker_index
					, void		*buffer
					, size_t	buffer_sz
					, void		*fldp
					, void		*user_cntxt
					)
{
	UTM_machine_t *internal = (UTM_machine_t *)user_cntxt;
	ssize_t	next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("prev_wrt0");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	PREV(internal);
	WRITE_0(fldp);
	internal->status = next_status;
	return	internal->status;
}


/* UTM function; PREV and WRITE CELL_1 */
static int prev_wrt1(size_t	 	checker_index
					, void		*buffer
					, size_t	buffer_sz
					, void		*fldp
					, void		*user_cntxt
					)
{
	UTM_machine_t *internal = (UTM_machine_t *)user_cntxt;
	ssize_t	next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("prev_wrt1");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	PREV(internal);
	WRITE_1(fldp);
	internal->status = next_status;
	return	internal->status;
}


/* UTM function; PREV and PREV */
static int prev_prev(size_t	 	checker_index
					, void		*buffer
					, size_t	buffer_sz
					, void		*fldp
					, void		*user_cntxt
					)
{
	UTM_machine_t *internal = (UTM_machine_t *)user_cntxt;
	ssize_t	next_status;

	/* test arguments and internal status */
	ILLEGAL_CALLBACKS_ARGS("prev_prev");

	next_status = (ssize_t)get_next_status((char *)fldp, internal->status);
	PREV(internal);
	PREV(internal);
	internal->status = next_status;
	return	internal->status;
}


/* get address of UTM function as the UTM function code */
static int (*utm_func_cd_2_func_addr[])(size_t	 	checker_index
										, void		*buffer
										, size_t	buffer_sz
										, void		*fldp
										, void		*user_cntxt
										) =
	{ term,      noop, wrt0,      wrt1,      next,      prev
	, wrt0_term, wrt0, wrt0,      wrt1,      wrt0_next, wrt0_prev
	, wrt1_term, wrt1, wrt0,      wrt1,      wrt1_next, wrt1_prev
	, next_term, next, next_wrt0, next_wrt1, next_next, noop
	, prev_term, prev, prev_wrt0, prev_wrt1, noop,      prev_prev
	};
#	define UTM_FUNC_CNT (sizeof(utm_func_cd_2_func_addr)				\
									/ sizeof(utm_func_cd_2_func_addr[0]))

#define UTM_FUNC_CD_2_FUNC_ADDR(UTM_FUNC_CD)							\
					(((UTM_FUNC_CD) < 0									\
					|| (UTM_FUNC_CD) >= (ssize_t)UTM_FUNC_CNT)?			\
						NULL: utm_func_cd_2_func_addr[(UTM_FUNC_CD)]	\
					)

/* print function pointer */
static void fprint_function_pointer(
								  FILE	*fp
								, int	(*funcp[])(size_t 	checker_index
												, void		*buffer
												, size_t	buffer_sz
												, void		*fldp
												, void		*user_cntxt
												)
								)
{
	size_t	i;
	char	*p = (char *)funcp;

	if (funcp == NULL) {
		fprintf(fp, "(nil)"); fflush(fp);
	}
	else {
		for (i = 0; i < sizeof(funcp); ++i) {
			printf("%02x", *p++);
		}
		fflush(fp);
	}

	return;
}

/* dump UTM function list */
static void dump_utm_func_list(int is_debug)
{
	if (is_debug) {
		ssize_t	i;

		for (i = 0; i < (ssize_t)UTM_FUNC_CNT; ++i) {
			printf("utm_func_cd_2_func_addr[%d] = ", i)
			; fprint_function_pointer(stdout, utm_func_cd_2_func_addr)
			; printf("\n"); fflush(stdout)
			;
		}
	}

	return;
}


/* skip spaces */
static char *skip_spaces(char *prog_code)
{
	while ((*prog_code != '\0') && isspace(*prog_code)) { ++prog_code;	}

	return	prog_code;
}


/* read the begin-of-list */
static char *read_begin_of_list(char *prog_code)
{
	prog_code = skip_spaces(prog_code); /* skip spaces */
	return	((*prog_code == '\0') || *prog_code++ != '(')? NULL: prog_code;
}


/* read the end-of-list */
static char *read_end_of_list(char *prog_code)
{
	prog_code = skip_spaces(prog_code); /* skip spaces */
	return	((*prog_code == '\0') || *prog_code++ != ')')? NULL: prog_code;
}


/* read the digit string */
static char *read_digit(char *prog_code, ssize_t *valp)
{
	char	*endp;

	prog_code = skip_spaces(prog_code);
	if ((*prog_code == '\0') || !isdigit(*prog_code)) {
		return NULL; /* lost the status-count */
	}

	/* convert the digit field */
	errno	= 0; /* initialize errno */
	*valp	= (ssize_t)strtol(prog_code, &endp, 10);
	if (((long)*valp == LONG_MAX || (long)*valp == LONG_MIN)
	&& errno == ERANGE
	) {	return	NULL; /* illegal format */	}

	return	endp; /* normal end */
}


/* read the UTM status description for CELL_0 or CELL_1 */
static char *read_status_desc(char		*prog_code
							, ssize_t	status_idx
							, ssize_t	match_side
							)
{
	ssize_t	utm_func_cd, next_status;

	/* search the begin of status-desc... */
	if ((prog_code = read_begin_of_list(prog_code)) == NULL) {
		return	NULL;	/* lost the parenthesis */
	}

	/* read the UTM function code */
	if ((((((prog_code = read_digit(prog_code, &utm_func_cd)) == NULL)
	|| *prog_code == '\0')
	|| !isspace(*prog_code))
	|| utm_func_cd < 0)
	|| utm_func_cd >= (ssize_t)UTM_FUNC_CNT
	) {	return	NULL; /* illegal UTM function code */	}

	/* read the next status */
	if ((((prog_code = read_digit(prog_code, &next_status)) == NULL)
	|| next_status < 0)
	|| next_status >= machine.status_cnt
	) {	return	NULL; /* illegal next status */	}

	/* search the end of status-desc... */
	if ((prog_code = read_end_of_list(prog_code)) == NULL) {
		return	NULL; /* lost parenthesis */
	}

	if (match_side == 0) {
		machine.matched_next_status[status_idx]		= next_status;
		machine.defines[status_idx].clause_cbs[0]
								= UTM_FUNC_CD_2_FUNC_ADDR(utm_func_cd);
	}
	else if (match_side == 1) {
		machine.unmatched_next_status[status_idx]	= next_status;
		machine.defines[status_idx].clause_cbs[1]
								= UTM_FUNC_CD_2_FUNC_ADDR(utm_func_cd);
		machine.defines[status_idx].clause_cbs[2]
								= NULL; /* no use this callback */
	}
	else {
		return	NULL;	/* illegal argument */
	}

	return	prog_code; /* normal end */
}

/* read the UTM program code, and build the Program_t structure */
static int read_program_code(char *prog_code)
{
	ssize_t	i;

	/* search the begin of program */
	if ((prog_code = read_begin_of_list(prog_code)) == NULL) {
		return	1; /* lost the begin of program */
	}

	/* read status-count... */
	if (((prog_code = read_digit(prog_code, &(machine.status_cnt)))
																== NULL)
	|| machine.status_cnt <= 0
	) {	return	2; /* illegal status-cnt */	}
	if ((*prog_code == '\0') || !isspace(*prog_code)) {
		return	3; /* illegal format */
	}

	/* allocate memory for `defines' and `next_status' */
	if ((((machine.defines = (Buffer_checker_t *)
						malloc(sizeof(*(machine.defines))
										* machine.status_cnt
										)) == NULL)
	|| (machine.matched_next_status = (ssize_t *)
						malloc(sizeof(*(machine.matched_next_status))
										* machine.status_cnt
										)) == NULL)
	|| (machine.unmatched_next_status = (ssize_t *)
						malloc(sizeof(*(machine.unmatched_next_status))
										* machine.status_cnt
										)) == NULL
	) {	return	4;	/* can't allocate memory */	}

	/* read status descriptions... */
	for (++prog_code, i = 0
	; (i < machine.status_cnt) && *prog_code != '\0'
	; ++i
	) {
		/* search the begin of status-desc... */
		if ((prog_code = read_begin_of_list(prog_code)) == NULL) {
			return	5; /* lost parenthesis */
		}

		if ((prog_code = read_status_desc(prog_code, i, 0)) == NULL) {
			return	6; /* illegal status desc for CELL_0 */
		}

		if ((prog_code = read_status_desc(prog_code, i, 1)) == NULL) {
			return	7; /* illegal status desc for CELL_1 */
		}

		/* search the end of status-desc... */
		if ((prog_code = read_end_of_list(prog_code)) == NULL) {
			return	8; /* lost parenthesis */
		}

		/* set the initial values the program */
		machine.defines[i].fld_offset			= 0;
		machine.defines[i].fldsz				= CELL_SZ;
		machine.defines[i].user_get_fld_addr	= get_cell_addr;
		machine.defines[i].regex				= CELL_0;
		machine.defines[i].preg					= NULL;
	}
	if (i != machine.status_cnt) {	return 9; /* lost status-desc */	}

	/* search the end of program... */
	if ((prog_code = read_end_of_list(prog_code)) == NULL) {
		return	10; /* lost the end of program */
	}

	return	0;
}


/* Program Entry */
int main(int ac, char **av)
{
	Buffer_checker_result_t	result;
	Regex_preg_list_t		regex_preg_list;
	ssize_t					tape_sz = 0L;
	size_t					executed_cnt = 0;
	char					*tape = NULL, *tp = NULL;
	int						i, exit_code = 1, res;

	/* test command options */
	if ((ac < 3) || strcmp(av[1], "-p") != 0) {
		usage(); exit(exit_code); return exit_code; /* Usage Error! */
	} ++exit_code;

	if (debug) {	dump_utm_func_list(debug);	}

	/* initialize regex_preg_list */
	memset(&regex_preg_list, (int)'\0', sizeof(regex_preg_list));

	/* initialize `program' according to `-p (prog)' option */
	if ((res = read_program_code(av[2])) != 0) {
		fprintf(stderr, "Illegal program (%d): \"%s %s\"\n"
				,				res, av[1], av[2]
				); fflush(stderr)
				;
		exit(exit_code); return exit_code; /* Illegal program */
	} ++exit_code;

	if (debug) { dump_program(&machine); printf("\n"); fflush(stdout); }

	/* initialize the tape according to command options... */
	for (i = 3; i < ac; ++i) {
		char	*endp;
		long	n;

		/* test result of `strtol(3)' */
		errno	= 0; /* initialize errno */
		n		= strtol((const char *)av[i], &endp, 10);
		if ((n == 0 && *endp != '\0')
		|| (errno == ERANGE && (n == LONG_MAX || n == LONG_MIN))
		) {
			fprintf(stderr, "\"%s\" is illegal number.\n"
					,	av[i]
					); fflush(stderr)
					;
			exit(exit_code); return exit_code; /* Illegal Number! */
		}

		/* allocate memory for the tape */
		if ((tape = (char *)realloc(tape, tape_sz + CELL_SZ * (n + 1)))
																== NULL
		) {
			fprintf(stderr
					, "Can't allocate memory for the %d-th arg "
									"\"%s\".\n"
					,				i
					,				av[i]
					); fflush(stderr)
					;
			exit(exit_code); return exit_code; /* Failed to realloc(3) */
		}

		/* set a positive number to the tape */
		tp		= tape + tape_sz;
		tape_sz	+= CELL_SZ * (n + 1);
		while (n-- > 0) { strncpy(tp, CELL_1, CELL_SZ); tp += CELL_SZ; }
		strncpy(tp, CELL_0, CELL_SZ); tp += CELL_SZ;
	} ++exit_code;

	/* append the end of tape (CELL_0) */
	tape_sz += CELL_SZ;
	if ((tape = (char *)realloc(tape, tape_sz)) == NULL) {
		fprintf(stderr
				, "Can't allocate memory for the end of tape "
								"(tape_sz = %ld).\n"
				,					(long)tape_sz
				); fflush(stderr)
				;
		exit(exit_code); return exit_code; /* Failed to realloc(3) */
	}
	strncpy(tape + (tape_sz - CELL_SZ), CELL_0, CELL_SZ);
	++exit_code;

	if (debug) {
		dump_buffer_checker_regex_preg_list(debug, &regex_preg_list)
		; fprintf(stderr, "\n"); fflush(stderr)
		;
	}

	/* initialize the `defines' */
	if ((result = init_buffer_checker(machine.defines
									, machine.status_cnt
									, tape_sz
									, &regex_preg_list
									)) != BCR_Continue
	) {
		fprintf(stderr
				, "!!FATAL!! init_buffer_checker(%p, %ld, %ld) returns "
					" %d (expected BCR_Continue = %d).\n"
				, (void *)(machine.defines), (long)machine.status_cnt
				,					(long)tape_sz
				,	result,			BCR_Continue
				); fflush(stderr)
				;
		goto	FATAL;
	} ++exit_code;

	if (debug) {
		dump_buffer_checker_regex_preg_list(debug, &regex_preg_list)
		; fprintf(stderr, "\n"); fflush(stderr)
		;
	}

	/* initialize the tape-head and machine status */
	machine.head = 0; machine.status = 0;

	if (debug) {
		printf(">>> initialized tape contains...\n")
		; fflush(stdout)
		;
	} dump_tape(tape, tape_sz); dump_internal(&machine);

	/* run the Turing-machine... */
	if ((result = exec_buffer_checker(tape
									, tape_sz
									, machine.defines
									, machine.status_cnt
									, &machine
									, &executed_cnt
									)) != BCR_Terminate
	) {
		fprintf(stderr, "UTM program -p '%s' stopped illegal "
				"status %d (expected = %d: BCR_Terminate).\n"
				,			av[2]
				,	result,		BCR_Terminate
				); fflush(stderr)
				;
		exit(exit_code); return exit_code;
	} ++exit_code;

	/* count up the cells those are set "1". */
	for (i = 0
	; strncmp((tape + (machine.head + CELL_SZ * i))
			, CELL_1
			, CELL_SZ) == 0
	; ++i) {	/* the last `i' means the counts of one-ed cells. */ ;	}

	/* print the sum of positive integers */
	printf("%d\n", i); fflush(stdout);

	if (debug) {
		fprintf(stderr
				, "=== exec_buffer_checker(%p, %ld, %p, %ld, "
								"%p, %p) "
				  "returns %d (BCR_Continue = %d, BCR_Terminate = %d, "
				  "BCR_Unmatched = %d). executed_cnt = %d.\n"
				, tape, (long)tape_sz, (void *)(machine.defines)
				,	(long)machine.status_cnt, (void *)&machine
				,		(void *)&executed_cnt
				,	result,		BCR_Continue, BCR_Terminate
				,			BCR_Unmatched,	executed_cnt
				); fflush(stderr)
				;
		printf("<<< calculated tape contains...\n"); fflush(stdout);
	} dump_tape(tape, tape_sz); dump_internal(&machine);

	/* release the `defines' allocated memories */
	if ((result = finish_buffer_checker(machine.defines
										, machine.status_cnt
										, &regex_preg_list
										)) != BCR_Continue
	) {
		fprintf(stderr
				, "!!FATAL!! finish_buffer_checker(%p, %ld) returns %d "
					"(expected BCR_Continue = %d).\n"
				, (void *)(machine.defines), (long)machine.status_cnt
				,												result
				,							BCR_Continue
				); fflush(stderr)
				;
	}
	else {
		if (debug) {
			dump_buffer_checker_regex_preg_list(debug, &regex_preg_list)
			; fprintf(stderr, "\n"); fflush(stderr)
			;
		}
		exit_code	= 0; /* Normal Exit-code */
	}

FATAL:
	/* release the tape allocated memories */
	free(tape); tape = NULL;

	exit(exit_code); return exit_code; /* Exit Program */
}
