/***********************************************************************
 * File: monetary.c
 * 	(Extracted from
 *	   "$Id: monetary.c,v 1.3 2009/05/30 20:52:31 maoyam Exp maoyam $")
 *
 * Abstracts:
 *   This C source implements the `monetary()' API that converts the
 * double value to a monetary formatted string.
 *
 * Copyright (C) 2005-2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#include "monetary.h"


/* This macro tests the digit specifications are illegal, or not. */
#define IS_ILLEGAL_DIGITS(sz,max_len,fractional_digits,additions)   	\
			(((((sz) - 1 < (max_len))				    				\
			|| (max_len) < (additions))				    				\
			|| ((additions) > 1 && (additions) <= (fractional_digits))) \
			|| ((additions) < 2 && (fractional_digits) != 0)			\
			)


/* This puts the comma at the triad digits. */
static void put_triad(char	*dstp
					, char	*dst_head
					, char	*srcp
					, char	*src_head
					)
{
	size_t	i;

	for (i = 0; (srcp >= src_head) && dstp >= dst_head; --srcp, ++i) {
		/* test: the 4th digit is numeric? */
		if (((isdigit(*srcp)) && i > 0) && i % 3 == 0) {
			*dstp--	= ','; /* put comma */

			/* test: reach to the head of destiny? */
			if (dstp < dst_head) {	break; /* filled the destiny */	}
		}
		*dstp--	= *srcp; /* put the digit or sign code */
	}

	return;
}

/*
 * This API converts the double value to a monetary string.
 *
 * arguments:
 *	amounts indicates the double value.
 *	max_len indicates the width of outputs.
 *	fractional_digits indicates the digits of decimal points.
 *	additions indicates the width of right hand of decimal points.
 *	str is pointing to formatted monetary string (output).
 *	sz indicates the size of str.
 *
 * returns:
 *	not NULL: the formatted monetary string (same as str).
 *	NULL 	: illegal arguments
 */
char	*monetary(double	amounts
				, size_t	max_len
				, size_t	fractional_digits
				, size_t	additions
				, char		*str
				, size_t	sz
				)
{
	size_t	i, len;
	char	*buf, *bp, *sp;

	errno	= 0; /* initialize errno */

	/* test: the arguments are illegal? */
	if ((str == NULL)
	|| IS_ILLEGAL_DIGITS(sz, max_len, fractional_digits, additions)
	) {	errno = EINVAL; return NULL; /* Failed: illegal arguments */	}

	/* convert the double value into a string */
	(void)snprintf(str, sz, "%*.*f"
					,		(int)max_len, (int)fractional_digits, amounts
					);

	/* test: `amounts' is not-a-number "NaN" or infinite ("-Inf","Inf")? */
	sp	= str + (strlen(str) - 1);
	if (!isdigit(*sp)) {
		errno = EDOM; return NULL; /* Failed: domain error by `amounts' */
	}

	/* allocate the formatting buffer */
	len	= max_len + 1;
	if ((buf = (char *)malloc(len)) == NULL) {
		errno = ENOMEM; return NULL; /* Failed: insufficient memory */
	}

	/* initialize the formatting buffer */
	memset(buf, (int)' ', len - 1); buf[len - 1] = '\0';

	/* set the integer parts pointer */
	if (fractional_digits > 0) {
		/* set the decimal part */
		for (i = 0
			, bp = buf + (len - 2 - (additions - fractional_digits - 1))
		; i <= fractional_digits
		; ++i
		) {	*bp-- = *sp--;	}
	}
	else {	/* set the decimal part */
		bp	= buf + (len - 2 - additions);
	}

	put_triad(bp, buf, sp, str); /* format the integer part */

	/* move the formatted string to the output */
	strcpy(str, buf + (len - 1 - max_len));

	free(buf); /* release the allocated formatting buffer */

	return	str; /* normal end */
}


/*
 * This API converts the numeric string to a monetary string.
 *
 * arguments:
 *	src is pointing to the numeric string.
 *	src_sz indicates the length of string that src is pointing to.
 *	max_len indicates the width of outputs.
 *	fractional_digits indicates the digits of decimal points.
 *	additions indicates the width of right hand of decimal points.
 *	str is pointing to formatted monetary string (output).
 *	sz indicates the size of str.
 *
 * returns:
 *	not NULL : the formatted monetary string (same as str).
 *	NULL : illegal arguments
 */
char	*strmonetary(char		*src
					, size_t	src_sz
					, size_t	max_len
					, size_t	fractional_digits
					, size_t	additions
					, char		*str
					, size_t	sz
					)
{
	char	*decimalp, *lastp, *dstp;

	errno	= 0; /* initialize errno */

	/* test: arguments are illegal? */
	if ((((src == NULL) || src_sz == 0) || str == NULL)
	|| IS_ILLEGAL_DIGITS(sz, max_len, fractional_digits, additions)
	) {	errno = EINVAL; return NULL; /* Failed: illegal arguments */	}

	/* skip the space codes... */
	for (decimalp = src
	; ((((size_t)(decimalp - src) < src_sz)
		&& *decimalp != '\0')
		&& isspace(*decimalp)
		)
	; ++decimalp
	) {	;	}

	/* test: the first non-space code isn't sign nor digit? */
	src	= decimalp;
	if (((*src == '\0')
	|| (((*src != '-') && *src != '+') && !isdigit(*src)))
	|| (((*src == '-') || *src == '+') && !isdigit(*(src + 1)))
	) {	errno = EINVAL; return NULL; /* Failed: illegal arguments */	}

	/* initialize the output string with space */
	memset(str, (int)' ', max_len); str[max_len] = '\0';

	/* set the integer part pointer */
	dstp	= str + (max_len - additions);

	/* find the decimal point... */
	for (decimalp = src + 1
	; (*decimalp != '\0') && isdigit(*decimalp)
	; ++decimalp
	) {	;	}

	/* test: the found decimal points? */
	if (((fractional_digits > 0)
	&& *decimalp == '.')
	&& isdigit(*(decimalp + 1))
	) {	/* put the decimal part */
		for (lastp = decimalp + 1
		; (*lastp != '\0') && isdigit(*lastp)
		; ++lastp
		) {	;	}
		memcpy(dstp
				, decimalp
				, (((size_t)(lastp - decimalp) > fractional_digits + 1)?
											fractional_digits + 1
							:				(size_t)(lastp - decimalp)
							)
				);
	}
	else {
		/* the src is pointing an integer string. */
		lastp	= decimalp;

		/* test: we are ordered to put the decimal part? */
		if (fractional_digits > 0) {
			/* put the decimal part of an integer string */
			*dstp = '.'; memset(dstp + 1, (int)'0', fractional_digits);
		}
	}

	--dstp; /* increment the output pointer */

	put_triad(dstp, str, decimalp - 1, src); /* format the integer part */

	return	str; /* normal end */
}


/*
 * This API convert the monetary string, that is output of `monetary()',
 * to the numeral string.
 *
 * arguments:
 *	mone points to the monetary string.
 *	dst points to buffer of numeral string outputs.
 *	sz indicates the size of the buffer pointed from `dst'.
 *
 * returns:
 *	NULL: occurred several errors, errno indicates details.
 *		EINVAL: illegal arguments
 *	otherwise: Normal end. converted string address (same as `dst')
 */
char *monetary_to_numeral(char *mone, char *dst, size_t sz)
{
#	define	copy_a_char(MONE,DST,SZ) {								\
					if ((SZ) < 2) { /* ERROR: sz is insufficient */	\
						errno = EINVAL; return NULL;				\
					}												\
					*DST++ = *MONE++; --SZ; /* transfer a char */	\
				}
	char	*result = dst;
	int		found_decimal_point = 0;

	/* test: illegal arguments? */
	if (((mone == NULL) || dst == NULL) || sz < 2) {
		errno = EINVAL; return NULL; /* ERROR: illegal args */
	}

	memset(dst, (int)'\0', sz); /* initialize the output buffer */

	/* skip the lead-off spaces... */
	while ((*mone != '\0') && *mone == ' ') { ++mone; }

	/* test: it has the negative or positive signs? */
	if (*mone == '-') {			copy_a_char(mone, dst, sz);			}
	else if (*mone == '+') {	++mone; /* skip positive sign */	}

	/* process the integer and fraction parts of `mone'... */
	while ((*mone != '\0') && *mone != ' ') {
		if (isdigit(*mone)) {
			/* *mone is digit character */
			copy_a_char(mone, dst, sz); /* copy it */
		}
		else if (*mone == ',') {
			/* *mone is triad delimiter */
			if (found_decimal_point != 0) {
				/* we processes fraction part, now. */
				errno = EINVAL; return NULL; /* ERROR */
			}
			++mone; /* skip it */
		}
		else if (*mone == '.') {
			/* mone is decimal point */
			if (found_decimal_point != 0) {
				/* we processes fraction part, now. */
				errno = EINVAL; return NULL; /* ERROR */
			}
			found_decimal_point	= !0;

			/* test: fraction part has several digits? */
			if ((*(mone + 1) != '\0')
			&& isdigit(*(mone + 1))
			) {	*dst++ = '.'; --sz; /* set comma */	}
			++mone; /* go to the next character */
		}
		else {
			/* *mone is not expected characters */
			errno = EINVAL; return NULL; /* ERROR: illegal arguments */
		}
	}

	return	result; /* normal end */
#	undef copy_a_char
}


/*
 * This API returns a double typed value such as the monetary string.
 *
 * arguments:
 *	mone is pointing to the monetary string.
 *
 * returns:
 *	errno indicates zero: normal end
 *	errno indicates non-zero: occurred several errors
 *		EINVAL: illegal arguments
 *		ERANGE: occurred overflow or underflow
 *		ENOMEM: insufficient memory
 */
double monetary_to_double(char *mone)
{
	double	result = 0.0;
	char	*numeral;
	size_t	sz;

	/* test: illegal arguments? */
	if (mone == NULL) {
		errno = EINVAL; return result; /* ERROR: result = 0.0 */
	}

	/* test: can't allocate the converted string area `numeral'? */
	if ((numeral = (char *)malloc((sz = strlen(mone) + 1))) == NULL) {
		/* insufficient memory */
		errno = ENOMEM; return result; /* ERROR: result = 0.0 */
	}

	/* test: converting monetary string is success? */
	memset(numeral, (int)'\0', sz);
	if (monetary_to_numeral(mone, numeral, sz) != NULL) {
		/* convert `numeral' string to the double typed value */
		char	*endptr = NULL;

		errno	= 0;
		result = strtod((const char *)numeral, &endptr);

		/* test: `numeral' isn't number? */
		if ((errno == 0) && numeral == endptr) {
			errno = EINVAL; /* ERROR: invalid arguments */
		}
	}

	/* release the allocated area `numeral' */
	free(numeral); numeral = NULL;

	return	result; /* If several error occurred, errno = EINVAL,
					 * ERANGE or EINVAL, and result = 0.0, HUGE_VAL
					 * or -HUGE_VAL.
					 * otherwise (normal end), errno is zero.
					 */
}


/*
 * This API rounds the double typed value.
 *
 * arguments:
 *	value indicates the double typed value.
 *	scale indicates the rounding digits.  if it's from zero to 17,
 *	this rounds fractional part.  if it's from -1 to -16, this
 *	rounds integer parts.
 *	how indicates EVEN, UP or DOWN for rounding value.
 *
 * returns:
 *	INF			: occurred several errors. errno indicates details.
 *		EINVAL	: illegal arguments
 *	otherwise	: rounded double value
 */
double round_double(double value, int scale, Round_double_how_t how)
{
	static const double		shifter[] = { 1.0
										, 10.0
										, 100.0
										, 1000.0
										, 10000.0
										, 100000.0
										, 1000000.0
										, 10000000.0
										, 100000000.0
										, 1000000000.0
										, 10000000000.0
										, 100000000000.0
										, 1000000000000.0
										, 10000000000000.0
										, 100000000000000.0
										, 1000000000000000.0
										, 10000000000000000.0
										, 100000000000000000.0
										};
	static const long long	llshifter[] =	{ 1LL
											, 10LL
											, 100LL
											, 1000LL
											, 10000LL
											, 100000LL
											, 1000000LL
											, 10000000LL
											, 100000000LL
											, 1000000000LL
											, 10000000000LL
											, 100000000000LL
											, 1000000000000LL
											, 10000000000000LL
											, 100000000000000LL
											, 1000000000000000LL
											, 10000000000000000LL
											};
	double					sign, result;

	/* test: illegal arguments? */
	if (((how != Round_double_EVEN) && how != Round_double_UP)
	&& how != Round_double_DOWN
	) {	errno = EINVAL; result = 1.0 / 0.0; /* illegal arguments */	}

	errno	= 0; /* initialize errno */
	sign	= (((fabs(value) == 0.0) || value > 0.0)? 1.0: -1.0);
	value	= fabs(value);

	if ((scale >= 0)
	&& scale < (int)(sizeof(shifter) / sizeof(shifter[0]))
	) {	/* round fractional part */
		result	= ((scale == 0)? value: value * shifter[scale]);
		if (how == Round_double_EVEN) {	result += 0.5;	}
		result	= ((how == Round_double_UP)?	ceil(result)
					:						floor(result)
					);
		if (scale != 0)	{	result /= shifter[scale];	}
		result	*= sign;
	}
	else if ((scale < 0)
	&& labs(scale) < (long)(sizeof(llshifter) / sizeof(llshifter[0]))
	) {	/* round integer part */
		long long ll = (long long)(floor(value));

		scale	*= -1;
		if (how == Round_double_EVEN) {	ll += (llshifter[scale] / 2LL);	}
		else if ((how == Round_double_UP)
		&& ((value > (double)ll) || ll % llshifter[scale] > 0)
		) {								ll += llshifter[scale];	}
		ll = ((ll / llshifter[scale]) * llshifter[scale]);
		result	= sign * (double)ll;
	}
	else {
		/* illegal argument `scale' */
		errno = EINVAL; result = 1.0 / 0.0; /* error : returns infinite */
	}

	return	result;
}
